package jalview.ws2.operations;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.function.Consumer;

import jalview.analysis.AlignSeq;
import jalview.analysis.AlignmentSorter;
import jalview.analysis.SeqsetUtils;
import jalview.analysis.SeqsetUtils.SequenceInfo;
import jalview.bin.Cache;
import jalview.datamodel.AlignedCodonFrame;
import jalview.datamodel.Alignment;
import jalview.datamodel.AlignmentI;
import jalview.datamodel.AlignmentOrder;
import jalview.datamodel.AlignmentView;
import jalview.datamodel.HiddenColumns;
import jalview.datamodel.Sequence;
import jalview.datamodel.SequenceI;
import jalview.gui.AlignViewport;
import jalview.util.MessageManager;
import jalview.ws.params.ArgumentI;
import jalview.ws2.WSJob;
import jalview.ws2.WSJobStatus;

public class AlignmentWorker extends AbstractPollableWorker
{
  private AlignmentOperation operation;
  
  private Consumer<AlignmentResult> resultConsumer;
  
  private AlignmentView msa;

  private AlignmentI dataset;

  private AlignViewport viewport;

  private List<AlignedCodonFrame> codonFrame = new ArrayList<>();

  private List<ArgumentI> args = Collections.emptyList();

  private boolean submitGaps = false;

  private boolean preserveOrder = false;

  private char gapCharacter;

  private WSJobList<AlignmentJob> jobs = new WSJobList<>();

  private Map<Long, Integer> exceptionCount = new HashMap<>();

  private static final int MAX_RETRY = 5;

  private static class JobInput
  {
    final List<SequenceI> inputSequences;

    final List<SequenceI> emptySequences;

    final Map<String, SequenceInfo> sequenceNames;

    private JobInput(List<SequenceI> inputSequences,
        List<SequenceI> emptySequences,
        Map<String, SequenceInfo> names)
    {
      this.inputSequences = Collections.unmodifiableList(inputSequences);
      this.emptySequences = Collections.unmodifiableList(emptySequences);
      this.sequenceNames = names;
    }

    boolean isInputValid()
    {
      return inputSequences.size() >= 2;
    }
  }
  
  public class AlignmentJob extends WSJob
  {
    private List<SequenceI> inputSequences;
    private List<SequenceI> emptySequences;
    private Map<String, SequenceInfo> sequenceNames;
    
    private AlignmentJob(String serviceProvider, String serviceName, 
        String hostName)
    {
      super(serviceProvider, serviceName, hostName);
    }
    
    private void setInput(JobInput input) {
      inputSequences = input.inputSequences;
      emptySequences = input.emptySequences;
      sequenceNames = input.sequenceNames;
    }
  }

  public class AlignmentResult
  {
    AlignmentI aln;

    List<AlignmentOrder> alorders;

    HiddenColumns hidden;

    AlignmentResult(AlignmentI aln, List<AlignmentOrder> alorders,
        HiddenColumns hidden)
    {
      this.aln = aln;
      this.alorders = alorders;
      this.hidden = hidden;
    }

    public AlignmentI getAlignment()
    {
      return aln;
    }

    public List<AlignmentOrder> getAlignmentOrders()
    {
      return alorders;
    }

    public HiddenColumns getHiddenColumns()
    {
      return hidden;
    }
  }

  public AlignmentWorker(AlignmentOperation operation,
      AlignmentView msa, List<ArgumentI> args,
      boolean submitGaps, boolean preserveOrder, AlignViewport viewport)
  {
    this.operation = operation;
    this.msa = msa;
    this.dataset = viewport.getAlignment().getDataset();
    List<AlignedCodonFrame> cf = Objects.requireNonNullElse(
        viewport.getAlignment().getCodonFrames(), Collections.emptyList());
    this.codonFrame.addAll(cf);
    this.args = args;
    this.submitGaps = submitGaps;
    this.preserveOrder = preserveOrder;
    this.viewport = viewport;
    this.gapCharacter = viewport.getGapCharacter();
  }

  @Override
  public Operation getOperation()
  {
    return operation;
  }

  @Override
  public WSJobList<? extends WSJob> getJobs()
  {
    return jobs;
  }
  
  public void setResultConsumer(Consumer<AlignmentResult> consumer)
  {
    this.resultConsumer = consumer;
  }
  
  @Override
  public void start() throws IOException
  {
    Cache.log.info(String.format("Starting new %s job.", operation.getName()));
    SequenceI[][] conmsa = msa.getVisibleContigs('-');
    if (conmsa == null)
    {
      return;
    }
    int numValid = 0;
    for (int i =  0; i < conmsa.length; i++)
    {
      JobInput input = prepareInputData(conmsa[i], 2, submitGaps);
      AlignmentJob job = new AlignmentJob(operation.service.getProviderName(),
          operation.getName(), operation.getHostName());
      job.setJobNum(i);
      job.setInput(input);
      jobs.add(job);
      listeners.fireJobCreated(job);
      if (input.isInputValid())
      {
        int count;
        String jobId = null;
        do
        {
          count = exceptionCount.getOrDefault(job.getUid(), MAX_RETRY);
          try
          {
            jobId = operation.service.submit(input.inputSequences, args);
            Cache.log.debug((String.format("Job %s submitted", job)));
            exceptionCount.remove(job.getUid());
          } catch (IOException e)
          {
            exceptionCount.put(job.getUid(), --count);
          }
        } while (jobId == null && count > 0);
        if (jobId != null)
        {
          job.setJobId(jobId);
          job.setStatus(WSJobStatus.SUBMITTED);
          numValid++;
        }
        else
        {
          job.setStatus(WSJobStatus.SERVER_ERROR);
        }
      }
      else
      {
        job.setStatus(WSJobStatus.INVALID);
        job.setErrorLog(
            MessageManager.getString("label.empty_alignment_job"));
      }
    }
    if (numValid > 0)
    {
      listeners.fireWorkerStarted();
    }
    else
    {
      listeners.fireWorkerNotStarted();
    }
  }
  

  private static JobInput prepareInputData(SequenceI[] sequences,
      int minLength, boolean submitGaps)
  {
    assert minLength >= 0 : MessageManager.getString(
        "error.implementation_error_minlen_must_be_greater_zero");
    int numSeq = 0;
    for (SequenceI seq : sequences)
    {
      if (seq.getEnd() - seq.getStart() >= minLength)
      {
        numSeq++;
      }
    }

    List<SequenceI> inputSequences = new ArrayList<>();
    List<SequenceI> emptySequences = new ArrayList<>();
    Map<String, SequenceInfo> names = new LinkedHashMap<>();
    for (int i = 0; i < sequences.length; i++)
    {
      SequenceI seq = sequences[i];
      String newName = SeqsetUtils.unique_name(i);
      var hash = SeqsetUtils.SeqCharacterHash(seq);
      names.put(newName, hash);
      if (numSeq > 1 && seq.getEnd() - seq.getStart() >= minLength)
      {
        String seqString = seq.getSequenceAsString();
        if (!submitGaps)
        {
          seqString = AlignSeq.extractGaps(
              jalview.util.Comparison.GapChars, seqString);
        }
        inputSequences.add(new Sequence(newName, seqString));
      }
      else
      {
        String seqString = "";
        if (seq.getEnd() >= seq.getStart()) // true if gaps only
        {
          seqString = seq.getSequenceAsString();
          if (!submitGaps)
          {
            seqString = AlignSeq.extractGaps(
                jalview.util.Comparison.GapChars, seqString);
          }
        }
        emptySequences.add(new Sequence(newName, seqString));
      }
    }

    return new JobInput(inputSequences, emptySequences, names);
  }

  @Override
  public void done()
  {
    listeners.fireWorkerCompleting();
    Map<Long, AlignmentI> results = new LinkedHashMap<>();
    for (WSJob job : getJobs())
    {
      if (job.getStatus().isFailed())
        continue;
      try
      {
        AlignmentI alignment = operation.getAlignmentSupplier().getAlignment(job);
        if (alignment != null)
        {
          results.put(job.getUid(), alignment);
        }
      } catch (Exception e)
      {
        if (!operation.getWebService().handleCollectionError(job, e))
        {
          Cache.log.error("Couldn't get alignment for job.", e);
          // TODO: Increment exception count and retry.
          job.setStatus(WSJobStatus.SERVER_ERROR);
        }
      }
    }
    if (results.size() > 0)
    {
      AlignmentResult out = prepareResult(results);
      resultConsumer.accept(out);
    }
    else
    {
      resultConsumer.accept(null);
    }
    listeners.fireWorkerCompleted();
  }

  private AlignmentResult prepareResult(Map<Long, AlignmentI> alignments)
  {
    List<AlignmentOrder> alorders = new ArrayList<>();
    SequenceI[][] results = new SequenceI[jobs.size()][];
    AlignmentOrder[] orders = new AlignmentOrder[jobs.size()];
    for (int i = 0; i < jobs.size(); i++)
    {
      AlignmentJob job = jobs.get(i);
      AlignmentI aln = alignments.get(job.getUid());
      if (aln != null) // equivalent of job.hasResults()
      {
        /* Get the alignment including any empty sequences in the original
         * order with original ids. */
        char gapChar = aln.getGapCharacter();
        List<SequenceI> emptySeqs = job.emptySequences;
        List<SequenceI> alnSeqs = aln.getSequences();
        // find the width of the longest sequence
        int width = 0;
        for (var seq : alnSeqs)
          width = Integer.max(width, seq.getLength());
        for (var emptySeq : emptySeqs)
          width = Integer.max(width, emptySeq.getLength());
        // pad shorter sequences with gaps
        String gapSeq = String.join("",
            Collections.nCopies(width, Character.toString(gapChar)));
        List<SequenceI> seqs = new ArrayList<>(
            alnSeqs.size() + emptySeqs.size());
        seqs.addAll(alnSeqs);
        seqs.addAll(emptySeqs);
        for (var seq : seqs)
        {
          if (seq.getLength() < width)
            seq.setSequence(seq.getSequenceAsString()
                + gapSeq.substring(seq.getLength()));
        }
        SequenceI[] result = seqs.toArray(new SequenceI[0]);
        AlignmentOrder msaOrder = new AlignmentOrder(result);
        AlignmentSorter.recoverOrder(result);
        Map<String, SequenceInfo> names = new HashMap<>(job.sequenceNames);
        // FIXME first call to deuniquify alters original alignment
        SeqsetUtils.deuniquify(names, result);
        alorders.add(msaOrder);
        results[i] = result;
        orders[i] = msaOrder;
      }
      else
      {
        results[i] = null;
      }
    }

    Object[] newView = msa.getUpdatedView(results, orders, gapCharacter);
    // free references to original data
    for (int i = 0; i < jobs.size(); i++)
    {
      results[i] = null;
      orders[i] = null;
    }
    SequenceI[] alignment = (SequenceI[]) newView[0];
    HiddenColumns hidden = (HiddenColumns) newView[1];
    Alignment aln = new Alignment(alignment);
    aln.setProperty("Alignment Program", operation.getName());
    if (dataset != null)
      aln.setDataset(dataset);

    propagateDatasetMappings(aln);
    return new AlignmentResult(aln, alorders, hidden);
    // displayNewFrame(aln, alorders, hidden);
  }

  /*
   * conserves dataset references to sequence objects returned from web
   * services. propagate codon frame data to alignment.
   */
  private void propagateDatasetMappings(Alignment aln)
  {
    if (codonFrame != null)
    {
      SequenceI[] alignment = aln.getSequencesArray();
      for (SequenceI seq : alignment)
      {
        for (AlignedCodonFrame acf : codonFrame)
        {
          if (acf != null && acf.involvesSequence(seq))
          {
            aln.addCodonFrame(acf);
            break;
          }
        }
      }
    }
  }
}
